"""Red Team Engine — public exports."""

from .types import (
    AttackCategory,
    AttackPayload,
    AttackOutcome,
    AttackResult,
    GuardrailEventCapture,
    RedTeamOptions,
    RedTeamProgress,
    RedTeamReport,
    CategoryScore,
    Vulnerability,
)
from .attacks import get_built_in_attacks, inject_system_prompt, BUILT_IN_ATTACKS
from .analyzer import analyze_attack_result, AnalysisInput, AnalysisResult
from .reporter import generate_report
from .runner import run_red_team
from .contextual_attacks import generate_contextual_attacks
from .regression import create_baseline, compare_reports, RedTeamBaseline, RegressionReport, CategoryRegression

__all__ = [
    "AttackCategory",
    "AttackPayload",
    "AttackOutcome",
    "AttackResult",
    "GuardrailEventCapture",
    "RedTeamOptions",
    "RedTeamProgress",
    "RedTeamReport",
    "CategoryScore",
    "Vulnerability",
    "get_built_in_attacks",
    "inject_system_prompt",
    "BUILT_IN_ATTACKS",
    "analyze_attack_result",
    "AnalysisInput",
    "AnalysisResult",
    "generate_report",
    "run_red_team",
    "generate_contextual_attacks",
    "create_baseline",
    "compare_reports",
    "RedTeamBaseline",
    "RegressionReport",
    "CategoryRegression",
]
